@@ -1,22 +1,24 @@
 #coding:utf-8
+from flask import current_app, url_for, request
+from flask_login import AnonymousUserMixin, current_user
 from user_agents import parse
 from blinker import Signal
-from flask import current_app, url_for, g, request, json, session
 from datetime import datetime, date
+from slugify import slugify 
+from loguru import logger
+from json import dumps
+from urllib.parse import unquote
 import random
 import string
 import dateutil.parser
-import traceback
-from flask_login import AnonymousUserMixin, current_user
 import hashlib
-import six
 import pycountry
-from slugify import slugify 
 import os
-from loguru import logger
-from json import dumps
 import hashlib
-from urllib.parse import unquote
+import shortuuid
+from pydantic import BaseModel, validator, EmailStr, conint, conlist, HttpUrl
+from typing import Optional, List
+
 
 class UndefinedEntityAttribute(Exception):
     pass
@@ -35,7 +37,7 @@ class BaseEntity(object):
         res = dict()
 
         # remove not schema attrs
-        for k, v in six.iteritems(doc):
+        for k, v in doc.items():
             if k in self._schemas.keys():
                 res[k] = v
 
@@ -47,7 +49,6 @@ class BaseEntity(object):
                 res[k] = self.__class__.format_attribute(v, self._schemas[k])
             except Exception as e:
                 logger.error('Attribute ({0}): {1}'.format(k, str(e)))
-                # print(traceback.print_exc())
                 raise e
 
         return res
@@ -83,6 +84,10 @@ class BaseEntity(object):
     def get_random_string(cls, _len):
         return "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(_len)) 
 
+    @classmethod
+    def get_urlsafe_string(cls, _len):
+        return shortuuid.ShortUUID().random(length=_len)
+
     @classmethod
     def format_attribute(cls, v, t):
         res = None
@@ -97,7 +102,7 @@ class BaseEntity(object):
                 if isinstance(v, (date, datetime)):
                     res = v
                 else:
-                    res = dateutil.parser.parse(v)
+                    res = dateutil.parser.isoparse(v)
         elif t == bool:
             res = bool(v)
         elif t == int:
@@ -133,10 +138,6 @@ class ElasticEntity(BaseEntity):
     def _table(cls):
         return "{0}.v1".format(cls.__name__.lower())
 
-    @classmethod
-    def _type(cls):
-        return '_doc'
-
     @classmethod
     def get(cls, _id=None, _count=10, _offset=0, _sort=[], _process=True, _source=None, _exclude=[], _all=False, _random=False, _range=None, **kwargs):
         if _id:
@@ -151,7 +152,7 @@ class ElasticEntity(BaseEntity):
         elif not _id:
             def pre_field(k, v):
                 return k if isinstance(v, bool) else '{0}.keyword'.format(k)
-            must = [{"terms" if isinstance(v, list) else "term": {pre_field(k, v): v}} for k, v in six.iteritems(kwargs)]
+            must = [{"terms" if isinstance(v, list) else "term": {pre_field(k, v): v}} for k, v in kwargs.items()]
 
             if _range:
                 _f, _s, _e = _range
@@ -219,7 +220,7 @@ class ElasticEntity(BaseEntity):
                     if _sid is None or len(resp['hits']['hits']) == 0:
                         break
 
-                    res += cls.process(resp['hits']['hits']) if _process else resp['hits']['hits']
+                    res += cls.process(resp['hits']['hits']) if _process else cls.format_source(resp['hits']['hits'])
                     found = resp['hits']['total']['value']
             else:
                 resp = current_app.es.search(index=cls._table(), body=q, request_timeout=60, ignore=[400, 404])
@@ -230,7 +231,7 @@ class ElasticEntity(BaseEntity):
 
                 logger.debug(u'RESPONSE took: {0}'.format(resp['took']))
 
-                res = cls.process(resp['hits']['hits']) if _process else resp['hits']['hits']
+                res = cls.process(resp['hits']['hits']) if _process else cls.format_source(resp['hits']['hits'])
                 found = resp['hits']['total']['value']
             return res, found
         return [], 0
@@ -246,6 +247,17 @@ class ElasticEntity(BaseEntity):
         logger.debug(u'END process result')
         return res
 
+    @classmethod
+    def format_source(cls, data):
+        res = []
+        for item in data:
+            for k, v in cls._schemas.items():
+                if k in item['_source'] and item['_source'][k]:
+                    if v == DateTimeField:
+                        item['_source'][k] = f"{item['_source'][k]}Z"
+            res.append(item)
+        return res
+
     @classmethod
     def put(cls, _id, doc, _refresh=True, _signal=True):
         obj = cls(doc, _id)
@@ -257,14 +269,16 @@ class ElasticEntity(BaseEntity):
         return resp, obj
 
     @classmethod
-    def delete(cls, _id):
-        resp = current_app.es.delete(index=cls._table(), id=_id, refresh=True, ignore=[404])
-        # cls.on_delete.send(response=resp)
+    def delete(cls, _id, _refresh=True, _signal=True):
+        resp = current_app.es.delete(index=cls._table(), id=_id, refresh=_refresh, ignore=[404])
+        if _signal:
+            cls.on_delete.send(response=resp)
         return resp
 
-    @classmethod
-    def process_response(cls, _items):
-        return [dict(item.to_dict(), **dict(_id=item._id)) for item in _items]
+    # @classmethod
+    # def process_response(cls, _items):
+    #     return [dict(item.to_dict(), **dict(_id=item._id)) for item in _items]
+
 
 class Page(ElasticEntity):
     on_put = Signal()
@@ -468,11 +482,8 @@ class Page(ElasticEntity):
         _countries.sort()
         _messages = {'w': _m1, 'b': _m2}
 
-        # logger.info(u'Response get_countries: {0}'.format(_messages))
-
         return _countries, _messages, len(_m1 + _m2) > 0
     
-
     @classmethod
     def get_options(cls):
         _cnt = 500
@@ -634,6 +645,7 @@ class Page(ElasticEntity):
             'countries': [{'iso': item.alpha_2, 'country': item.name} for item in pycountry.countries],
             'countries_name': [item.name for item in pycountry.countries],
             'services': ['casino', 'poker', 'betting', 'lotto', 'bingo'],
+            'status': ['draft', 'on_review', 'published'],
             'category': ['provider', 'slot', 'page', 'collection'],
             'collection_category': ['provider', 'slot'],
             'wager_types': ['xb', 'x(d+b)'],
@@ -759,32 +771,10 @@ class Page(ElasticEntity):
             return current_user.country_full in self.geo
         return True
 
-    @property
-    def toc(self):
-        # search by content (h2, h3)
-        return [('item1', 'Item 1'), ('item2', 'Item 2')]
-
-    @property
-    def toc_content(self):
-        # processed content (add h2 anchors, converts links, custom snippets)
-        return self.content
-
     @property
     def activity_str(self):
         return self.activity_count or 0
 
-    @property
-    def software_str(self):
-        return 'NetEnt (+12)'
-
-    @property
-    def licence_str(self):
-        return 'MGA (+1)'
-
-    @classmethod
-    def slots(cls, _exclude=[], _count=12, _source=["title", "alias", "cover", "software"], **params):
-        return [], 0
-
     @classmethod
     def provider_by_context(cls, country, _exclude=[], _count=12, _page=1, _source=["title", "alias", "logo"], _aggs={}, _sorting=None, _locale='en', **params):
         filter_list =  [ 
@@ -917,6 +907,7 @@ class Page(ElasticEntity):
         else:
             return cls.process(resp['hits']['hits']), resp['hits']['total']['value']
 
+
 class Guest(AnonymousUserMixin):
     @property
     def is_admin(self):
@@ -984,7 +975,7 @@ class Actor(ElasticEntity, Guest):
     
     @classmethod
     def _table(cls):
-        return "{0}_actor.v1".format(os.environ.get('PROJECT', 'project'))
+        return "{0}_actor.{1}".format(os.environ.get('PROJECT', 'project'), os.environ.get('VERSION', 'v1'))
 
     @property
     def is_authenticated(self):
@@ -1038,4 +1029,363 @@ class Activity(ElasticEntity):
         
     @classmethod
     def _table(cls):
-        return "{0}_activity.v1".format(os.environ.get('PROJECT', 'project'))
+        return "{0}_activity.{1}".format(os.environ.get('PROJECT', 'project'), os.environ.get('VERSION', 'v1'))
+
+
+class TdsHit(ElasticEntity):
+    on_put = Signal()
+
+    _schemas = {
+        'createdon': DateTimeField,
+        'stream': str,
+        'stream_name': str,
+        'campaign_name': str,
+        'campaign_alias': str,
+        'campaign_id': str,
+        'click_id': str,
+        'ip': str,
+        'country': str,
+        'country_iso': str,
+        'ua': str,
+        'is_bot': bool,
+        'is_uniq': bool,
+        'action': str,
+        'url': str,
+        'subid': str,
+        'fingerprint': str,
+    }
+        
+    @classmethod
+    def _table(cls):
+        return "{0}_tds_hits.{1}".format(os.environ.get('PROJECT', 'project'), os.environ.get('VERSION', 'v1'))
+
+    @classmethod
+    def aggs(cls, _field, _range=None, _cnt=500):
+
+        def get_subaggs(_a={}):
+            _b = {
+                "fingerprint": {
+                    "cardinality": {
+                        "field": "fingerprint.keyword"
+                    }
+                },
+                "is_bot": {
+                    "terms": {
+                        "field": "is_bot",
+                    }
+                }
+            }
+            if _a:
+                _b = dict(_b, **_a)
+            return _b
+
+        if _field in ['createdon']:
+            aggs = {
+                _field: {
+                    "date_histogram": {
+                        "field": _field,
+                        "calendar_interval": "1d"
+                    },
+                    "aggs": get_subaggs()
+                },
+            }        
+        elif _field in ['country']:
+            aggs = {
+                _field: {
+                    "terms": {
+                        "field": f"{_field}.keyword",
+                        "size": _cnt
+                    },
+                    "aggs": get_subaggs({
+                        "country_iso": {
+                            "terms": {
+                                "field": "country_iso.keyword"
+                            },
+                        }
+                    })
+                },
+            }        
+        elif _field in ['campaign_id']:
+            aggs = {
+                _field: {
+                    "terms": {
+                        "field": f"{_field}.keyword",
+                        "size": _cnt
+                    },
+                    "aggs": get_subaggs({
+                        "campaign_name": {
+                            "terms": {
+                                "field": "campaign_name.keyword"
+                            },
+                        }
+                    })
+                },
+            }        
+        else:
+            aggs = {
+                _field: {
+                    "terms": {
+                        "field": f"{_field}.keyword",
+                        "size": _cnt
+                    },
+                    "aggs": get_subaggs()
+                },
+            }        
+
+        must = []
+
+        if _range:
+            _f, _s, _e = _range
+            must.append({"range": {_f: {"gte": _s, "lte": _e}}})
+
+        q = {
+            "query": {
+                "bool": {
+                    "must": must
+                }
+            },
+            "size": 0,
+            "aggs": aggs
+        }
+
+        resp = current_app.es.search(index=cls._table(), body=q, request_timeout=60, ignore=[400, 404])
+        logger.info(f'Response: {resp}')
+
+        aggs = {}
+        if 'status' in resp and resp['status'] in [400, 404]:
+            pass
+        else:
+            aggs = resp.get('aggregations', [])            
+
+        logger.info(f'Aggs found: {aggs}')
+
+        def get_uc(item):
+            return item['fingerprint']['value']
+
+        def get_bots(item):
+            _v = [item['doc_count'] for item in item['is_bot']['buckets'] if item['key_as_string'] == 'true']
+            return _v.pop() if _v else 0
+
+        def process(items, key):
+            if key in items:
+                if key == 'createdon':
+                    return [{'term': item['key_as_string'].strip(), 'hits': item['doc_count'], 'uc': get_uc(item), 'bots': get_bots(item)} for item in items[key]['buckets']]
+                elif key == 'country':
+                    return [{'term': item['key'].strip(), 'country_iso': item['country_iso']['buckets'][0]['key'].lower(), 'hits': item['doc_count'], 'uc': get_uc(item), 'bots': get_bots(item)} for item in items[key]['buckets']]
+                elif key == 'campaign_id':
+                    return [{'term': ', '.join([k['key'] for k in item['campaign_name']['buckets']]), 'hits': item['doc_count'], 'uc': get_uc(item), 'bots': get_bots(item)} for item in items[key]['buckets']]
+                else:
+                    return [{'term': item['key'].strip(), 'hits': item['doc_count'], 'uc': get_uc(item), 'bots': get_bots(item)} for item in items[key]['buckets']]
+            return []
+
+        return process(aggs, _field), 0 #resp['hits']['total']['value']
+
+
+class TdsDomain(ElasticEntity):
+    on_put = Signal()
+
+    _schemas = {
+        'createdon': DateTimeField,
+        'domain': str,
+        'endpoint': str,
+        'is_https': bool,
+    }
+        
+    @classmethod
+    def _table(cls):
+        return "{0}_tds_domain.{1}".format(os.environ.get('PROJECT', 'project'), os.environ.get('VERSION', 'v1'))
+
+
+class TdsCampaign(ElasticEntity):
+    on_put = Signal()
+
+    _schemas = {
+        'createdon': DateTimeField,
+        'updatedon': DateTimeField,
+        'name': str,
+        'domain': str,
+        'alias': str,
+        'groups': list,
+        'ttl': int,
+        'is_split': bool,
+        'is_active': bool,
+        'is_archive': bool,
+        'notes': str,
+        'streams': list,
+    }
+        
+    @classmethod
+    def _table(cls):
+        return "{0}_tds_campaign.{1}".format(os.environ.get('PROJECT', 'project'), os.environ.get('VERSION', 'v1'))
+
+
+# pydantic models
+
+class UserModel(BaseModel):
+    login: str = ...
+
+    @validator('login', allow_reuse=True)
+    def fn_not_empty(cls, v):
+        assert v != '', 'Required field'
+        return v
+
+
+class ManagerModel(BaseModel):
+    login: str = ...
+    skype: str = ...
+
+    @validator('login', 'skype', allow_reuse=True)
+    def fn_not_empty(cls, v):
+        assert v != '', 'Required field'
+        return v
+
+
+class SignupModel(BaseModel):
+    login: str = ...
+    skype: str = ...
+    invite_code: str = ...
+
+    @validator('login', 'skype', allow_reuse=True)
+    def fn_not_empty(cls, v):
+        assert v != '', 'Required field'
+        return v
+
+    @validator('invite_code', allow_reuse=True)
+    def fn_invite_code(cls, v):
+        def is_valid(s):
+            if len(s) == 56:
+                objs, total = Page.get(_id=s)
+                if total == 1:
+                    obj = objs.pop()
+                    if obj.category == 'provider':
+                        return True
+            return False
+        
+        assert v != '', 'Required field'
+        assert v == current_app.config['DASHBOARD_INVITE'] or is_valid(v), 'Invite Expired'
+        return v
+
+
+class AuthModel(BaseModel):
+    login: str = ...
+    password: str = ...
+
+    @validator('login', 'password', allow_reuse=True)
+    def fn_not_empty(cls, v):
+        assert v != '', 'Required field'
+        return v
+
+
+class CasinoModel(BaseModel):
+    @validator('*', pre=True, allow_reuse=True)
+    def empty_str_to_none(cls, v):
+        if v == '':
+            return None
+        return v
+
+    @validator('year', pre=True, allow_reuse=True)
+    def year_validator(cls, v):
+        if v:
+            s = str(v)
+            if s.isdigit() and (int(s) < 1990 or int(s) > datetime.utcnow().year):
+                raise ValueError('Incorrect value')
+        return v        
+
+    title: str = ...
+    website: HttpUrl = ...
+    languages: List[str] = []
+    services: List[str] = []
+    year: Optional[int]
+    support_languages: List[str] = []
+    support_livechat: bool = False
+    support_email: Optional[EmailStr] = None
+    support_phone: Optional[str] = None
+    attachs: List[dict] = []
+    logo: Optional[str] = None
+    logo_small: Optional[str] = None
+    licences: List[str] = []
+    games: List[str] = []
+    games_count: Optional[int]
+    software: List[str] = []
+    currencies: conlist(str, min_items=1)
+    default_currency: str
+    deposits: List[str] = []
+    min_deposit: int
+    kyc_deposit: bool = False
+    withdrawal: List[str] = []
+    min_withdrawal: int
+    withdrawal_monthly: Optional[int] = None
+    kyc_withdrawal: bool = False
+    welcome_package_max_bonus: Optional[int] = None
+    welcome_package_match: Optional[int] = None
+    welcome_package_fs: Optional[int] = None
+    welcome_package: str
+    welcome_package_note: Optional[str] = None
+    welcome_package_wager: Optional[int] = None
+    cashback: Optional[int] = None
+    geo_whitelist: Optional[str] = None
+    geo_blacklist: Optional[str] = None
+    geo: List[str] = []
+    content: Optional[str] = None
+
+
+class DomainModel(BaseModel):
+    @validator('*', pre=True, allow_reuse=True)
+    def empty_str_to_none(cls, v):
+        if v == '':
+            return None
+        return v
+
+    domain: str = ...
+    endpoint: str = ...
+    is_https: bool = False
+
+
+class CampaignModel(BaseModel):
+    @validator('*', pre=True, allow_reuse=True)
+    def empty_str_to_none(cls, v):
+        if v == '':
+            return None
+        return v
+
+    name: str = ...
+    domain: str = ...
+    alias: str = ...
+    groups: Optional[List[str]] = []
+    ttl: int
+    is_split: bool = False
+    is_active: bool = False
+    is_archive: bool = False
+    notes: Optional[str] = None
+    streams: conlist(dict, min_items=1)
+
+
+class ActivityTicketModel(BaseModel):
+    subject: str = ...
+    contacts: str = ...
+    name: str = ...
+    message: str = ...
+
+    @validator('subject', 'contacts', 'name', 'message', allow_reuse=True)
+    def not_empty(cls, v):
+        assert v != '', 'Required field'
+        return v
+
+
+class ActivitySubscribeModel(BaseModel):
+    email: EmailStr
+    is_agree: bool = False
+
+    @validator('is_agree', pre=True, always=True, allow_reuse=True)
+    def check_agree(cls, v):
+        assert v == True, 'You must accept our terms'
+        return v   
+
+
+class ActivityFeedbackModel(BaseModel):
+    rate: conint(ge=1, le=5)
+    casino: str = ...
+    casino_id: str = ...
+    name: Optional[str]
+    pros: Optional[str]
+    cons: Optional[str]                
\ No newline at end of file
